import subprocess
import os


def concat_all_ts(video_file: str) -> None:
    """
    Concat all ts chucks one by one.

    Parameters
    ----------
    video_file: str
        The file_name by which to save the concatenated .ts files
    """
    subprocess.Popen(["ffmpeg", "-f", "concat", "-safe", "0", "-i", "ts_list.txt",
                      "-c", "copy", f"{video_file}.ts"], stdout=subprocess.DEVNULL).wait()
    os.unlink("ts_list.txt")


def convert_video(video_input: str, video_output: str) -> None:
    """
    Convert video from ts to mp4.

    Parameters
    ----------
    video_input : str
        The input file name

    video_output : str
        The output file name
    """
    flags = ["ffmpeg", "-i", f"{video_input}.ts", "-acodec"
        , "copy", "-vcodec", "copy", video_output]
    subprocess.Popen(flags, stdout=subprocess.DEVNULL).wait()
    os.unlink(f"{video_input}.ts")


def get_ts_start_time(file_path: str) -> float:
    """
    Search For ts file start time.

    Parameters
    ----------
    file_path : str
        The file_path to the file to be probed

    Returns
    -------
    float
        The start time of the .ts file located at file_path
    """
    parse_png_to_mpeg2ts_stream(file_path)
    return float(os.path.split(file_path)[-1])


def parse_png_to_mpeg2ts_stream(file_path: str) -> None:
    """
    Strip the mpeg2-ts data from a file.

    Parameters
    ----------
    file_path: str
        The file_path to be converted to mpeg2-ts, note this function assumes that the file will
        be detected as png due to starting bytes, and re parse it as mpeg2-ts
    """
    cmd = "ffmpeg -y -f mpegts -i {} -c copy {}.ts -v quiet".format(file_path, file_path)
    subprocess.Popen(cmd, shell=True).wait()
    os.unlink(file_path)
    os.rename(f"{file_path}.ts", file_path)



from multiprocessing import current_process
from traceback import print_exc
from queue import Queue
from typing import Any

from .common.constants import PORT, HEADER_SIZE,IP
from .common.base import Server

import platform
import pickle
import socket
import sys


def send_data(client: socket.socket, data: str) -> None:
    """
    Abstract the sending of data to the client.

    Parameters
    ----------
    client : socket.socket
        The client socket to which we send data

    data : str
        The data to send to the client socket
    """
    client.send(bytes(data, "utf-8"))
    client.close()


def receive_data(client: socket.socket, d_type: str = "str", debug: bool = False) -> Any:
    """
    Abstract receiving data from client.

    Parameters
    ----------
    client : socket.socket
        The client socket which will receive data

    d_type : str, optional
        A string specifying the expected data type (default is "str")

    debug : bool
        A bool specifying whether to print debug options (default is False)

    Returns
    -------
    str
        The data received from the client socket
    """
    got_data = []
    while True:
        temp = client.recv(4096)
        if len(temp) <= 0 or not temp:
            break
        got_data.append(temp)

    data = b"".join(got_data)

    client.close()

    if debug:
        print(f"Data received by server of size {len(data)} bytes")

    if d_type == "bytes":
        return pickle.loads(data)
    return data.decode("utf-8")


class ProducerServerProcess:
    """
    Producer Server Process.

    A class that models a server that acts as a way of passing messages between download_process and
    video_handling process.

    Attributes
    ----------
    __server : Server
        The socket that listens on the ip and port provided in the __init__() method

    __queue : Queue
        The worker queue that is responsible for passing data between download_process and video_handling process

    __stop : bool
        This variable is used to tell the producer_server_process to stop the server

    __sent : int
        This is used to keep track of the items sent to the video_handling process

    Methods
    -------
    start()
        This method is used to start the ProducerServerProcess server

    process_action(action: str, client: socket.socket)
        This method is used to process the actions received by the download_process and video_handling process
    """

    def __init__(self, ip: str, port: int):
        """
        Initialize object of ProducerServerProcess to ip and port parameters.

        Parameters
        ----------
        ip : str
            The ip address to bind the server with.

        port : int
            The port to bind the server with.
        """
        self.__server: Server = Server(ip, port)
        self.__queue: Queue = Queue()
        self.__update_links: int = 0
        self.__stop_queue = False
        self.__stop: bool = False
        self.__sent: int = 0

    def start(self, debug: bool) -> None:
        """Start the server."""
        soc = self.__server.socket
        while not self.__stop:
            try:

                client, address = soc.accept()
                data = client.recv(HEADER_SIZE)
                action = data.decode("utf-8").split()[0].strip()
                if debug:
                    print(f"Received action {action} from {address}")
                self.process_action(action, client)

            except (KeyboardInterrupt, Exception):
                print_exc()
                sys.exit()

    def process_action(self, action: str, client: socket.socket) -> None:
        """
        Perform actions depending on the flow of file.

        Parameters
        ----------
        action : str
            The action that the server needs to perform

        client : socket.socket
            The socket connected with the client
        """
        # The POST_FILENAME_QUEUE action is used to add a file_name to the work queue, from the
        # download_process
        if action == "POST_FILENAME_QUEUE":
            data = receive_data(client, "bytes")
            for item in data:
                self.__queue.put(item)

        elif action == "STOP_QUEUE":
            data = receive_data(client).strip()
            self.__stop_queue = True
            self.__update_links = int(data)

        # The GET_FILENAME_QUEUE action is used to query the work queue to get the next task to the
        # video_handling process
        elif action == "GET_FILENAME_QUEUE" and not self.__queue.empty():
            self.__sent += 1
            data = self.__queue.get()
            send_data(client, data)

        elif action == "GET_FILENAME_QUEUE" and self.__stop_queue:
            send_data(client, str(self.__update_links))

        elif action == "STOP":
            self.__stop = True


# producer_server_process is responsible for passing of messages between the download_process and
# video_handling process supported actions: - POST_FILENAME_QUEUE, GET_FILENAME_QUEUE, POST_CONVERT_VIDEO,
# GET_CONVERT_VIDEO, POST_CONCAT_TS, GET_CONCAT_TS, STOP
def producer_server_process(debug=False):
    """Start the ProducerServerProcess."""
    print(f"Started Producer Process {current_process().name}")

    if platform.system() == "Windows":
        producer = ProducerServerProcess(IP, PORT)
    else:
        producer = ProducerServerProcess("", PORT)

    try:

        producer.start(debug)

    except (KeyboardInterrupt, Exception):
        print_exc()
        sys.exit()

    print(f"Stopped Producer Server {current_process().name}")



import pathlib
from typing import List, Dict
from urllib.parse import urlparse, urljoin

import requests


def fetch_playlist_links(session: requests.Session, playlist_url: str,
                         keep: bool = False) -> List[str]:
    """Fetch the m3u8 playlist from the playlist_url.

    Parameters
    ----------
    session : requests.Session
        The session via which the request will be made
    playlist_url : str
        The url where the m3u8 playlist is hosted
    keep : bool
        Keep file with links on disk.

    Returns
    -------
    List[str]
        A list with all the urls where the .ts files are hosted
    """
    filename = "links.txt"
    p = pathlib.Path(playlist_url)
    if p.is_file():
        with open(playlist_url) as file:
            temp = file.readlines()
        links: List[str] = [link.strip() for link in temp if not link.startswith("#")]
    else:
        res: requests.Response = session.get(playlist_url, timeout=60)
        temp = [link.strip() for link in res.text.splitlines()]
        parsed_url = urlparse(playlist_url)
        base_url: str = f"{parsed_url.scheme}://{parsed_url.netloc}{'/'.join(parsed_url.path.split('/')[:-1])}/"
        # Construct a links list containing the links joined with the base_url.
        # If the link in the m3u8 playlist is already a full link we add that to the result `links list`
        # else we join the base_url with the link partial in the playlist and add that to the `links list`
        links: List[str] = [(link if parsed_url.scheme in link else urljoin(base_url, link))
                            for link in temp if not link.startswith("#")]
    if keep:
        with open(filename, "w") as file:
            for link in links:
                file.write(f"{link}\n")

    return links


def construct_file_name_links_map(links: List[str]) -> Dict[str, str]:
    """Construct a dictionary with link to file name mappings.

    Parameters
    ----------
    links : List[str]
        A list with all the urls of the hosted .ts files

    Returns
    -------
    Dict[str, str]
        A dictionary containing link as key and the corresponding file name as value
    """
    link_file_map = {}

    for file_name, link in enumerate(links):
        link_file_map[link] = str(file_name)

    return link_file_map



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait
from multiprocessing import current_process, JoinableQueue
from typing import List, Dict, Optional
from hyper.contrib import HTTP20Adapter
from traceback import print_exc
from queue import Queue
from math import ceil
from time import time
from copy import copy

from .common.constants import IP, PORT, HEADER_SIZE
from .weblib.fetch import fetch_data
from .common.base import Client

import platform
import requests
import pickle
import sys
import os


def download_process(links, total_links, session, http2, max_retries,
                     convert, file_link_maps, path_prefix, debug, progress_bar_queue,
                     processes, threads) -> None:
    print(f"Starting Download process {current_process().name}")
    start_time = time()
    try:
        download_manager = DownloadProcess(links, total_links, session, http2,
                                           max_retries, convert, debug, processes, threads)

        start_processes(download_manager, file_link_maps, path_prefix, progress_bar_queue)
        try:
            client = Client(IP, PORT)
            client.send_data(f"{'STOP_QUEUE':<{HEADER_SIZE}}{download_manager.get_total_downloaded_links_count()}")
        except (ConnectionRefusedError, ConnectionResetError):
            print_exc()

    except (KeyboardInterrupt, Exception):
        print_exc()

    minutes, seconds = divmod(time() - start_time, 60)
    print(f"\nDownload took {minutes} minutes and {seconds} seconds")
    print(f"Stopped Download process {current_process().name}")


class DownloadProcess:
    def __init__(self, links: List[str], total_links: int, session: requests.Session,
                 http2: bool = False, max_retries: int = 5,
                 convert: bool = True, debug: bool = False,
                 processes: int = 0, threads: int = 0):
        """Initialize Object of DownloadProcess.

        Parameters
        ----------
        links: List[str]
            The links to be downloaded
        total_links: int
            The number of links to be downloaded
        session: requests.Session
            The session using which all the links will be downloaded
        http2: bool
            A flag to keep track of the type of http headers
        max_retries: int
            The maximum number of retries for a particular link
        convert: bool
            A flag to keep track of the whether the downloaded video should be converted
        debug: bool
            A flag to print messages to the console
        processes: int
            The number of processes to be used
        threads: int
            The number of threads to be used by each process
        """
        self.__session: requests.Session = session
        self.__total_links: int = total_links
        self.__links: List[str] = links
        self.max_retries: int = max_retries
        self.http2: bool = http2
        self.convert = convert
        self.__sent = 0
        self.__process_num = processes or (len(os.sched_getaffinity(os.getpid())) if platform.system() == "Linux" else 4)
        self.__thread_num = threads or int(ceil((total_links - self.__sent) / (self.__process_num * 4)))
        self.debug = debug
        self.done_retries = 0
        self.error_links = []

    def get_thread_num(self) -> int:
        return self.__thread_num

    def get_process_num(self) -> int:
        return self.__process_num

    def set_thread_num(self, val: int) -> None:
        self.__thread_num = val

    def get_download_links(self) -> List[str]:
        return self.__links

    def get_total_links(self) -> int:
        return self.__total_links

    def get_session(self) -> requests.Session:
        return self.__session

    def get_total_downloaded_links_count(self) -> int:
        return self.__sent

    def set_total_downloaded_links_count(self, val: int) -> None:
        self.__sent = val


def start_processes(download_manager: DownloadProcess, file_link_maps: Dict[str, str], path_prefix: str
                    , progress_bar_queue) -> None:
    process_num: int = download_manager.get_process_num()
    if download_manager.debug:
        print(f"starting {process_num} processes for {len(download_manager.get_download_links())} links")

    with ProcessPoolExecutor(max_workers=process_num) as process_pool_executor:
        try:
            process_pool_executor_handler(process_pool_executor, download_manager, file_link_maps, path_prefix
                                          , progress_bar_queue)
        except (KeyboardInterrupt, Exception):
            sys.exit()


def process_pool_executor_handler(executor: ProcessPoolExecutor, manager: DownloadProcess,
                                  file_maps: Dict[str, str], directory: str, progress_bar_queue) -> None:
    done_queue = JoinableQueue()

    def update_hook(future: Future):
        temp = future.result()
        if temp:
            for failed_links in temp:
                done_queue.put(failed_links)

    while manager.done_retries != manager.max_retries:
        print(f"Starting download {manager.get_total_links() - manager.get_total_downloaded_links_count()} links left")
        available_cpus = list(os.sched_getaffinity(os.getpid())) if platform.system() == "Linux" else [0, 1, 2, 3]
        print(f"available cpu's {available_cpus}, initializing {4 * manager.get_process_num()}"
              f" threads with {manager.get_thread_num()} links per "
              f"process")

        if len(manager.error_links):
            download_links = manager.error_links.copy()
            manager.error_links = []
        else:
            download_links = manager.get_download_links().copy()

        process_futures: List[Future] = []

        start = 0
        for temp_num in range(len(download_links)):
            end = start + manager.get_thread_num()

            if end > len(download_links):
                end = len(download_links)

            cpu_num = available_cpus[temp_num % len(available_cpus)]
            process_futures.append(executor.submit(start_threads, download_links[start:end],
                                                   file_maps, manager.get_session(), directory,
                                                   manager.http2, progress_bar_queue, manager.debug, cpu_num))
            process_futures[-1].add_done_callback(update_hook)
            start = end

            if end >= len(download_links):
                break

        wait(process_futures)

        while not done_queue.empty():
            link = done_queue.get()
            manager.error_links.append(link)

        manager.set_total_downloaded_links_count(manager.get_total_links() - len(manager.error_links))

        if manager.debug:
            print(f"Total downloaded links {manager.get_total_downloaded_links_count()}")
            print(f"Error links generated {len(manager.error_links)}")

        if len(manager.error_links):
            manager.set_thread_num(int(ceil((manager.get_total_links()
                                             - manager.get_total_downloaded_links_count())
                                            / manager.get_process_num())))
            print(f"\n{manager.get_total_links()} was expected but "
                  f"{manager.get_total_downloaded_links_count()} was downloaded.")
            manager.done_retries += 1
            print(f"Trying retry {manager.done_retries}")
        else:
            break


def start_threads(links: List[str], maps: Dict[str, str], session: requests.Session,
                  file_path_prefix: str, http2: bool, progress_bar_queue, debug: bool = False,
                  cpu_num: int = 0) -> List[Optional[str]]:
    failed_links = Queue()
    sessions_queue = Queue()

    def update_hook(future: Future):
        temp = future.result()
        if temp:
            failed_links.put(temp)

    sent_links = {}
    if platform.system() == "Linux":
        os.sched_setaffinity(os.getpid(), {cpu_num})

    THREAD_WORKERS: int = 4

    if http2:
        for _ in range(THREAD_WORKERS):
            sess = requests.Session()
            sess.mount("https://", HTTP20Adapter(max_retries=10))
            sess.headers = copy(session.headers)
            sessions_queue.put(sess)

    with ThreadPoolExecutor(max_workers=THREAD_WORKERS) as executor:
        for link in links:
            temp_path = os.path.join(file_path_prefix, maps[link])
            sent_links[link] = temp_path

            if http2:
                new_session = None
            else:
                new_session = session

            thread_future = executor.submit(download_thread, temp_path, link, new_session,
                                            http2, sessions_queue)
            thread_future.add_done_callback(update_hook)

    failed = []

    for link in failed_links.queue:
        del sent_links[link]
        failed.append(link)

    send_data = pickle.dumps(list(sent_links.values()))
    msg = f"{'POST_FILENAME_QUEUE':<{HEADER_SIZE}}"
    client = Client(IP, PORT)
    client.send_data(msg)
    client.send_data(send_data, "bytes")

    if debug:
        print(f"Sending data to server of size {len(send_data)} bytes")

    progress_bar_queue.put(len(sent_links))
    return failed


def download_thread(file_path: str, link: str, session: requests.Session,
                    http2: bool, s_queue: Queue) -> Optional[str]:
    if os.path.exists(file_path):
        return None

    if not session and http2:
        session = s_queue.get()

    done_url = fetch_data(link, session, 120, file_path, http2)

    if http2 and session:
        s_queue.put(session)

    return done_url



from tqdm import tqdm
from time import sleep

def update_progress_bar(queue, length):
    with tqdm(total=length) as pbar:
        while True:
            try:
                data = queue.get()
                sleep(0.1)
                pbar.update(data)
            except EOFError:
                pass



from requests.adapters import HTTPAdapter
from hyper.contrib import HTTP20Adapter
from multiprocessing import Process, Manager
from traceback import print_exc
from shutil import rmtree

from .m3u8lib.parser import fetch_playlist_links, construct_file_name_links_map
from .producer_server_process import producer_server_process
from .video_handling_process import video_handling
from .download_process import download_process
from .weblib.parse import construct_headers
from .progressbar import update_progress_bar

import platform
import requests
import argparse
import sys
import os


def directory_validator(string: str) -> str:
    # This code is used to check if passed in headers_path string is valid path to a file
    # if not FileNotFoundError is raised
    if os.path.isfile(string):
        return string
    raise FileNotFoundError(f"{string} does not point to a file")


def main():
    # start the program with -h or --help to get more info on how to use the script.
    parser = argparse.ArgumentParser()
    parser.add_argument("url", help="Pass in a url containing m3u8 playlist", type=str)
    parser.add_argument("-n", "--name", type=str, help="Specify a name to save the downloaded video as, if no name is "
                                                       "specified default name of 'video' will be chosen")
    parser.add_argument("-p", "--header-path", type=directory_validator,
                        help="Specify the path to the file containing headers, if no path is specified the program "
                             "will look for a headers.txt file in the same directory")
    parser.add_argument("-r", "--retry", type=int, help="Specify number of retries by default 5 retries will be "
                                                        "initiated")
    parser.add_argument("-f", "--force", action="store_true", help="If this flag is used and the video has been "
                                                                   "downloaded the download will restart")
    parser.add_argument("-c", "--convert", help="Convert the downloaded video to mp4 using ffmpeg", action="store_true")
    parser.add_argument("-d", "--debug", help="Print helpful messages to the terminal to "
                                              "help understanding the process flow", action="store_true")
    parser.add_argument("-m", "--processes", type=int, help="Specify number of processese by default 4 will be "
                                                        "initiated, or number of CPUs if on Linux")
    parser.add_argument("-t", "--threads", type=int, help="Specify number of threads by default 4 will be "
                                                        "initiated for each process")

    cli_args = parser.parse_args()

    # HTTP/1.1 adapter
    ADAPTER1: HTTPAdapter = HTTPAdapter(max_retries=5)

    # HTTP/2 adapter
    ADAPTER2: HTTP20Adapter = HTTP20Adapter(max_retries=10)

    url = cli_args.url

    if cli_args.name:
        name = cli_args.name
    else:
        name = "video"

    if cli_args.header_path:
        path = cli_args.header_path
    else:
        path = "headers.txt"

    MAX_RETRIES = 5
    if cli_args.retry:
        MAX_RETRIES = cli_args.retry

    headers, http2 = construct_headers(path)
    debug = cli_args.debug

    # Mount new connection adapters to the session created.
    sess: requests.Session = requests.Session()
    sess.mount("http://", ADAPTER1)
    if http2:
        sess.mount("https://", ADAPTER2)

    sess.headers = headers

    links = fetch_playlist_links(sess, url, debug)
    file_link_maps = construct_file_name_links_map(links)
    path_prefix = "".join([i for i in url if i.isalnum()])

    if platform.system() == "Windows":
        path_prefix = path_prefix[:10]
    else:
        path_prefix = "." + path_prefix

    os.makedirs(path_prefix, exist_ok=True)

    try:
        server = Process(target=producer_server_process, args=(debug,), name="producer_server_process")
        server.start()

        video = Process(target=video_handling, args=(len(links), name, cli_args.convert, debug),
                        name="video_handling_process")
        video.start()

        queue = Manager().Queue()

        progress_bar_process = Process(target=update_progress_bar, args=(queue, len(links)),
                                       name="progress_bar_process_going_on")

        progress_bar_process.daemon = True
        progress_bar_process.start()

        download_process(links, len(links), sess, http2, MAX_RETRIES, cli_args.convert,
                         file_link_maps, path_prefix, debug, queue, cli_args.processes, cli_args.threads)

        server.join()
        video.join()

        rmtree(path_prefix)
    except (KeyboardInterrupt, Exception):
        print_exc()

    sys.exit()



from typing import Dict, Tuple
from pprint import pprint
from time import sleep

import sys
import os


def construct_headers(header_path: str) -> Tuple[Dict[str, str], bool]:
    """
    Construct headers from Header file.

    Parameters
    ----------
    header_path: str
        The path to the file containing the headers

    Returns
    -------
    Tuple[Dict[str, str], bool]
        Returns a tuple containing the headers and http2 variable
    """
    http2 = False
    header: Dict[str, str] = {}

    print("Parsed Request headers\n")

    if os.path.exists(header_path):
        with open(header_path) as file:
            lines = file.readlines()
            if not lines:
                lines.append("Access-Control-Allow-Origin: *")
                lines.append("User-Agent: Mozilla/5.0 (X11; Linux x86_64) "
                             "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36")

            lines = [line.strip() for line in lines]

        for line in lines:
            if line[0] == ":":

                # This code checks if the given headers are of HTTP/2, if yes then HTTP2 is set to True
                # and all subsequent requests will be made with HTTP/2

                temp = line.split(":")
                if temp[1] and temp[2:]:
                    header[":" + temp[1]] = ":".join(temp[2:]).strip()
                    http2 = True
            else:
                temp = line.split(":")
                if temp[0] and temp[1:]:
                    header[temp[0]] = ":".join(temp[1:]).strip()

        if "cookie" in header:
            del header["cookie"]

        pprint(header)
        print(f"press ctrl+c or ctrl+z if parsed headers of type http2 {http2} are incorrect")

        try:
            sleep(10)
        except KeyboardInterrupt:
            print("headers incorrectly parsed check construct_headers() function for debugging")
            sys.exit()
    else:
        header["Access-Control-Allow-Origin"] = "*"
        header["User-Agent"] = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 " \
                               "(KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36"

    return header, http2



from typing import Optional, Any
import socket


class BaseSocketServer:

    """
    A class used to abstract the sending and receiving of data via a socket.

    ...

    Attributes
    ----------
    socket : socket.socket
        A socket object with which the abstraction needs to be made.

    Methods
    -------
    send_data(data: Any)
        A method to send data to the connected socket.

    receive_data(size: int = 4096)
        A method to receive data from the connected socket.
    """

    def __init__(self, soc: socket.socket):
        """
        Initialize BaseSocketServer Object to soc.

        Parameters
        ----------
        soc : socket.socket
            A socket object with which the abstraction needs to be made.
        """
        self.socket = soc

    def send_data(self, data: Any, d_type: str = "str") -> None:
        """
        Send data to the connected socket.

        Parameters
        ----------
        data : Any
            The data to be sent to the connected socket.

        d_type : str
            The type of data to be sent
        """
        if d_type == "bytes":
            self.socket.send(data)
        else:
            self.socket.send(bytes(data, "utf-8"))

    def receive_data(self, size: int = 4096, data_type: str = str) -> Optional[Any]:
        """
        Receive data to the connected socket.

        Parameters
        ----------
        size : int, optional
            The size of data the socket should accept (default is 4096).

        data_type : str, optional
            The expected data_type of the received data (default is str).

        Returns
        -------
        Optional[Any]
            The data received by the socket
        """
        data = self.socket.recv(size)
        if len(data) <= 0 or data is None:
            return None
        if data_type == bytes:
            return data

        return data.decode("utf-8")


class Server(BaseSocketServer):

    """
    A class used to abstract the socket that will act as the server.

    ....
    """

    def __init__(self, ip, port):
        """
        Initialize Server Object with ip and port parameters.

        Parameters
        ----------
        ip : str
            A string specifying the ip to bind the socket with.

        port: int
            A number specifying the port to bind the socket with.
        """
        super(Server, self).__init__(socket.socket(
            socket.AF_INET, socket.SOCK_STREAM))
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((ip, port))
        self.socket.listen(10)


class Client(BaseSocketServer):

    """
    A class used to abstract the socket that will connect with the server.

    ...
    """

    def __init__(self, ip, port):
        """
        Initialize Client Object with ip and port parameters.

        Parameters
        ----------
        ip : str
            A string specifying the ip to bind the socket with.

        port: int
            A number specifying the port to bind the socket with.
        """
        super(Client, self).__init__(socket.socket(
            socket.AF_INET, socket.SOCK_STREAM))
        self.socket.connect((ip, port))



from urllib.parse import urlparse
from typing import Optional

import requests


def fetch_data(download_url: str, session: requests.Session,
               timeout: int, file_path: str, http2: bool) -> Optional[str]:
    """
    Fetch Data from Url.

    Parameters
    ----------
    download_url : str
        The url that hosts the .ts file

    session : requests.Session
        The session using which we can make a request

    timeout : int
        The time period after which the request expires

    file_path : str
        The path where the file needs to be stored

    http2: bool
        A boolean to specify the need for http2 support

    Returns
    -------
    Optional[str]
        Returns a string containing the download link that failed
    """
    try:
        if http2:
            if ":path" in session.headers:
                parsed_suffix = urlparse(download_url).path
                session.headers[":path"] = parsed_suffix

        with session.get(download_url, timeout=timeout, stream=True) as request_data:
            request_data.raise_for_status()
            if request_data.status_code == 302:
                request_data = redirect_handler(session, request_data)

            file = open(file_path, "ab")

            for chunk in request_data.iter_content(10485760):
                if not chunk:
                    break

                file.write(chunk)
            
            file.close()

    except (ConnectionResetError, ConnectionRefusedError, ConnectionError,
            TimeoutError, ConnectionAbortedError, OSError):
        return download_url


def redirect_handler(session: requests.Session, request_body: requests.Response, retry: int = 5) -> requests.Response:
    text = request_body.content.decode().split(" ")
    url = text[-1]
    parsed_url = urlparse(url)

    authority = parsed_url.netloc
    path = parsed_url.path

    temp_auth = session.headers[":authority"]
    temp_path = session.headers[":path"]
    temp_origin = session.headers["origin"]

    session.headers[":authority"] = authority
    session.headers[":path"] = path
    session.headers["origin"] = "null"

    request_data = session.get(url, stream=True)

    session.headers[":authority"] = temp_auth
    session.headers[":path"] = temp_path
    session.headers["origin"] = temp_origin

    if request_data.status_code == 403:
        raise ConnectionAbortedError("403")
    elif request_data.status_code == 302 and retry:
        return redirect_handler(session, request_data, retry - 1)

    return request_data



from setuptools import setup, find_packages

DESCRIPTION = 'm3u8 playlist downloader'
VERSION = "0.5.0"


with open("README.md") as fh:
    LONG_DESCRIPTION = fh.read()

setup(
    name="m3u8dl",
    version=VERSION,
    zip_safe=False,
    author="Kevin Rohan Vaz",
    author_email="excalibur.krv@gmail.com",
    maintainer='Kevin Rohan Vaz',
    maintainer_email='excalibur.krv@gmail.com',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/excalibur-kvrv/m3u8-dl",
    packages=find_packages(),
    install_requires=[
        "hyper==0.7.0",
        "requests==2.26.0",
        "tqdm==4.62.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from multiprocessing import current_process
from traceback import print_exc
from random import SystemRandom
from time import sleep

from .videolib.convertor import convert_video, concat_all_ts, get_ts_start_time
from .common.constants import PORT,IP
from .common.base import Client

import platform
import sys


def get_task(total_links, file_meta_data, stop=False, debug=False):
    while not stop:
        try:
            while True:
                if platform.system() == "Windows":
                    client = Client(IP, PORT)
                else:
                    client = Client("", PORT)

                client.send_data("GET_FILENAME_QUEUE")
                data = client.receive_data()
                if debug:
                    print(f"begin processing {data}")
                if data:
                    break
                timer = SystemRandom()
                sleep(timer.randrange(1, 10))

            if data.isnumeric():
                if debug:
                    print(f"Processed links {len(file_meta_data)} links downloaded by server {data}")
                break

            if data and not data.isnumeric():
                file_meta_data[data] = get_ts_start_time(data)

            if len(file_meta_data) == total_links:
                stop = True
                continue

            elif not data.isnumeric():
                continue
        except:
            print_exc()
            sys.exit()

    return file_meta_data


def start_process(total_links, file_name, convert, debug):
    file_meta_data = get_task(total_links, {}, False, debug)

    with open("ts_list.txt", "w") as file:
        for file_path in sorted(list(file_meta_data.keys()), key=lambda k: file_meta_data[k]):
            file.write(f"file '{file_path}'\n")

    if platform.system() == "Windows":
        client = Client(IP, PORT)
    else:
        client = Client("", PORT)

    client.send_data("STOP")

    if debug:
        print("Concatenating")
    concat_all_ts(file_name)

    if convert:
        if debug:
            print("Converting to mp4")
        convert_video(file_name, f"{file_name}.mp4")


# video_handling process is responsible for constructing a graph to filter out duplicate video frames,
# concat all .ts files and convert the resulting concatenation to mp4
def video_handling(total_links, file_name, convert, debug=False):
    print(f"Starting video handling process {current_process().name}")

    try:
        sleep(10)
        start_process(total_links, file_name, convert, debug)

    except (KeyboardInterrupt, Exception):
        print_exc()
        sys.exit()

    print(f"Stopping video handling process {current_process().name}")
